/*
 * Copyright (C) 2008 Fabien Chereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _SKYBACKGROUND_HPP_
#define _SKYBACKGROUND_HPP_

#include "StelModule.hpp"
#include <QString>
#include <QMap>

class StelCore;
class SkyImageTile;
class QProgressBar;

//! Manage the sky background images, including DSS and deep sky objects images
class SkyBackground : public StelModule
{
	Q_OBJECT
			
public:
	SkyBackground();
	~SkyBackground();

	///////////////////////////////////////////////////////////////////////////
	// Methods defined in the StelModule class
	//! Initialize
	virtual void init();
	
	//! Draws sky background
	virtual void draw(StelCore* core);
	
	//! Update state which is time dependent.
	virtual void update(double deltaTime) {;}
	
	//! Update i18
	virtual void updateI18n() {;}
	
	//! Determines the order in which the various modules are drawn.
	virtual double getCallOrder(StelModuleActionName actionName) const;

	///////////////////////////////////////////////////////////////////////////
	// Other specific methods
	//! Add a new sky image tile in the list of background images
	QString insertSkyImage(SkyImageTile* img, bool show=true, bool externallyOwned=true);
	
	//! Remove a sky image tile from the list of background images
	void removeSkyImage(SkyImageTile* img);
	
public slots:
	///////////////////////////////////////////////////////////////////////////
	// Properties setters and getters
	//! Set whether Sky Background should be displayed
	void setFlagShow(bool b) {flagShow = b;}
	//! Get whether Sky Background should be displayed
	bool getFlagShow() const {return flagShow;}
	
	///////////////////////////////////////////////////////////////////////////
	// Other slots
	//! Add a new image from its URI (URL or local file name)
	//! The image is owned by the manager and will be destroyed at the end of the program
	//! or when removeSkyImage is called with the same URI
	//! @param uri the local file or the URL where the JSON image description is located
	//! @param show defined whether the image should be shown by default
	//! @return the reference key to use when accessing this image later on
	QString insertSkyImage(const QString& uri, bool show=true);
	
	//! Remove a sky image tile from the list of background images
	//! @param key the reference key generated by insertSkyImage
	void removeSkyImage(const QString& key);
	
private slots:
	//! Called when loading of data started or stopped for one collection
	//! @param b true if data loading started, false if finished
	void loadingStateChanged(bool b);
	
	//! Called when the percentage of loading tiles/tiles to be displayed changed for one collection
	//! @param percentage the percentage of loaded data
	void percentLoadedChanged(int percentage);
	
private:
	class SkyBackgroundElem
	{
		public:
			SkyBackgroundElem(SkyImageTile* t, bool show=true, bool externallyOwned=true);
			~SkyBackgroundElem();
			SkyImageTile* tile;
			QProgressBar* progressBar;
			bool show;
			bool externallyOwned;
	};
	
	SkyBackgroundElem* skyBackgroundElemForTile(const SkyImageTile*);
	
	QString keyForTile(const SkyImageTile*);
	
	//! Map image key/image
	QMap<QString, SkyBackgroundElem*> allSkyImages;
	
	// Whether to draw at all
	bool flagShow;
};

#endif // _SKYBACKGROUND_HPP_
